# # coding: utf-8
# from icu import RuleBasedCollator
# from pathlib import Path
#
#
# rules = Path(__file__).parent / "../third_party/rules.txt"
# collator = RuleBasedCollator(
#     "[normalization on]\n[reorder Tibt]\n" + rules.read_text(encoding="utf-8")
# )
#
#
# def bo_sorted(word_list):
#     return sorted(word_list, key=collator.getSortKey)



from pathlib import Path
from shutil import rmtree

import click
from bordr import __version__ as bordr__version
from botok import Text, WordTokenizer
from botok import __version__ as botok__version__
from botok import expose_data
from pyewts import VERSION as pyewts__version__

from pybo import __version__ as pybo__version__
from pybo.corpus.parse_corrected import extract_new_entries
from pybo.pipeline.pipes import pybo_form, pybo_mod, pybo_prep
from pybo.rdr.rdr import rdr as r
from pybo.rdr.rdr_2_replace_matcher import rdr_2_replace_matcher
from pybo.utils.profile_report import profile_report as p_report
from pybo.utils.regex_batch_apply import batch_apply_regex, get_regex_pairs


@click.group()
@click.version_option(pybo__version__)
def cli():
    pass


@cli.command()
def info():
    click.echo("pybo install path: " + str(Path(__file__).parent.resolve()))
    click.echo("pybo: " + pybo__version__)
    click.echo("botok: " + botok__version__)
    click.echo("pyewts: " + pyewts__version__)
    click.echo("bordr: " + bordr__version)


def prepare_folder(main=None, custom=None, overwrite=False):
    profile = "POS"
    # 1. MAIN PROFILE
    if not main:
        # for better Windows support:
        # https://stackoverflow.com/questions/6227590/finding-the-users-my-documents-path/6227623#6227623
        main = Path.home() / "Documents/pybo/main"
    else:
        main = Path(main)
    main.mkdir(parents=True, exist_ok=True)

    if overwrite:
        rmtree(main, ignore_errors=True)
        main.mkdir()

    try:
        expose_data(main, profile=profile)
    except IOError:
        click.echo('using the existing data in "Documents/pybo/main/"')

    # 2. CUSTOM PROFILE
    if not custom:
        custom = Path.home() / "Documents/pybo/custom"
    else:
        custom = Path(custom)
    custom.mkdir(exist_ok=True)
    for dir in ["adjustment", "remove", "words", "words_skrt"]:
        Path(custom / dir).mkdir(exist_ok=True)

    return main, custom


# Tokenize file
@cli.command()
@click.argument("input-dir", type=click.Path(exists=True))
@click.option(
    "-t",
    "--tags",
    help="""Select and order the tags. Available tags are:
t-clean_text, p-pos, l-lemma, s-sense.\n
Usage: `-t tpl` will give for every token `<raw-text>/<clean-text>/<pos>/<lemma>`
and will give just `<raw-text>` if tag option is not specified.""",
)
@click.option("-o", type=click.Path(exists=True))
@click.option("-p", type=click.Path(exists=True), help="main-profile path")
@click.option(
    "-p2",
    multiple=True,
    type=(click.Path(exists=True), click.Path(exists=True)),
    help="paths: main-profile, custom-profile",
)
@click.option("-w", "--overwrite", is_flag=True)
@click.option("-r", "--rebuild-trie", is_flag=True)
def tok(**kwargs):
    input_dir = Path(kwargs["input_dir"])
    if kwargs["o"] is not None:
        output_dir = Path(kwargs["o"])
    else:
        output_dir = input_dir.parent / (input_dir.name + "_tok")
        output_dir.mkdir(exist_ok=True)
    p = kwargs["p"]
    p2 = kwargs["p2"]
    overwrite = kwargs["overwrite"]
    rebuild = kwargs["rebuild_trie"]

    # prepare folder folder to receive all the botok files
    if p and p2:
        click.echo(
            "Choose either -p or -p2 for the tokenizer's profiles, not both\nExiting"
        )
        exit(1)

    if p:
        main, custom = prepare_folder(p, overwrite=overwrite)
        click.echo("main profile: " + p)
    elif p2:
        main, custom = prepare_folder(p2[0][0], p2[0][1], overwrite=overwrite)
        click.echo("main/custom profiles: " + str(p2))
    else:
        main, custom = prepare_folder(overwrite=overwrite)
        click.echo("using default profile")

    wt = WordTokenizer(
        tok_profile=main,
        tok_modifs=custom,
        tok_mode="custom",
        adj_profile=main,
        adj_modifs=custom,
        adj_mode="custom",
        conf_path=main.parent,
        build_trie=rebuild,
    )

    def pybo_tok(in_str):
        return wt.tokenize(in_str)

    # Select and Order the tags
    if kwargs["tags"]:
        pybo_mod.__defaults__ = (list(kwargs["tags"]),)

    for f in input_dir.glob("*.txt"):
        out_file = output_dir / (f.stem + "_tok.txt")
        text = Text(f, out_file)
        text.custom_pipeline(pybo_prep, pybo_tok, pybo_mod, pybo_form)


# Tokenize string
@cli.command()
@click.argument("string")
def tok_string(**kwargs):
    t = Text(kwargs["string"])
    click.echo(t.tokenize_words_raw_lines)


# create report for botok profiles
@cli.command()
@click.argument("profile", type=click.Path(exists=True))
def profile_report(**kwargs):
    p_report(kwargs["profile"])


# rdr_2_replace_matcher
@cli.command()
@click.argument("infile", type=click.Path(exists=True))
def rdr2repl(**kwargs):
    infile = Path(kwargs["infile"])
    outfile = infile.parent / (infile.stem + ".yaml")
    dump = infile.read_text(encoding="utf-8-sig")
    processed = rdr_2_replace_matcher(dump)
    outfile.write_text(processed, encoding="utf-8-sig")


# # sort in the Tibetan order
# @cli.command()
# @click.argument("infile", type=click.Path(exists=True))
# def kakha(**kwargs):
#     infile = Path(kwargs["infile"])
#     words = infile.read_text(encoding="utf-8-sig").split()
#     words = bo_sorted(words)
#     infile.write_text("\n".join(words), encoding="utf-8-sig")


# generate rdr rules
@cli.command()
@click.argument("input", type=click.Path(exists=True))
@click.option("-o", "--out-dir", type=click.Path(exists=True))
@click.option("-k", "--keep", type=str)
def rdr(**kwargs):
    file_or_dir = Path(kwargs["input"])
    keep = "none" if kwargs["keep"] is None else kwargs["keep"]
    out_dir = Path(kwargs["out_dir"]) if kwargs["out_dir"] else None

    log = None
    if file_or_dir.is_dir():
        file = file_or_dir / (file_or_dir.name + "_rules")
        with open(file, encoding="utf-8-sig", mode="w") as tmp:
            # concatenate all the content of input
            for f in file_or_dir.glob("*.txt"):
                tmp.write(f.read_text(encoding="utf-8-sig") + " ")
        log = r(file, outdir=out_dir, keep=keep)
        file.unlink()
    elif file_or_dir.is_file():
        log = r(file_or_dir, kwargs["out_dir"], keep=keep)
    else:
        click.echo(f'"{file_or_dir}" does not exist.')

    click.echo(log)


# extract new entries from manually corrected texts + existing profile
@cli.command()
@click.argument("corrected-path", type=click.Path(exists=True))
@click.argument("profile-path", type=click.Path(exists=True))
@click.option("-o", "--out-dir", type=click.Path(exists=True))
def profile_update(**kwargs):
    corrected = Path(kwargs["corrected_path"])
    profile = Path(kwargs["profile_path"])
    out_dir = Path(kwargs["out_dir"]) if kwargs["out_dir"] else None

    dump = ""
    for f in corrected.glob("*.txt"):
        dump += f.read_text(encoding="utf-8-sig") + "\n"

    rules = extract_new_entries(dump, profile)
    if not out_dir:
        out = corrected.parent / (corrected.name + "_words.tsv")
    else:
        out = out_dir / (corrected.name + "_words.tsv")

    if not out.parent.is_dir():
        out.parent.mkdir(exist_ok=True)

    out.write_text(rules, encoding="utf-8-sig")


# FNR - Find and Replace with a list of regexes
@cli.command()
@click.argument("in-dir", type=click.Path(exists=True))
@click.argument("regex-file", type=click.Path(exists=True))
@click.option("-o", "--out-dir", type=click.Path(exists=True))
@click.option("-t", "--tag")
def fnr(**kwargs):
    # get the args
    indir = Path(kwargs["in_dir"])
    regex_file = Path(kwargs["regex_file"])
    out_dir = Path(kwargs["out_dir"]) if kwargs["out_dir"] else None

    if not indir.is_dir():
        click.echo("IN_DIR should be a folder, not a file.\nexiting...")
        exit(1)

    # optional out file tag
    tag = kwargs["tag"] if kwargs["tag"] else regex_file.stem

    # generate rules
    rules = get_regex_pairs(regex_file.open(encoding="utf-8-sig").readlines())

    # apply on each file, prefixing each one with the regex filename
    for f in indir.rglob("*.txt"):
        if not f.stem.startswith("_"):
            string = f.read_text(encoding="utf-8-sig")
            out = batch_apply_regex(string, rules)
            name = f"_{tag}_" + f.name
            if out_dir:
                Path(out_dir).mkdir(parents=True, exist_ok=True)
                outfile = out_dir / name
            else:
                outfile = f.parent / name
            outfile.write_text(out, encoding="utf-8-sig")


if __name__ == "__main__":
    cli()



# coding: utf-8
from pathlib import Path
from collections import defaultdict


def profile_entries(pathname):
    pathname = Path(pathname)
    entries = defaultdict(list)

    profile_files = [Path(__file__).parent.parent / "resources/particles.tsv"]
    for d in pathname.glob("*"):
        # filter unwanted directories and files
        dirs_ignored = ["adjustment", "entry_data"]
        if not d.is_dir() or d.name in dirs_ignored or d.name.startswith("."):
            continue

        profile_files.extend(list(d.glob("*.tsv")))

    # add files
    for f in profile_files:
        lines = f.read_text(encoding="utf-8-sig").splitlines()
        for num, line in enumerate(lines):
            if line.startswith("#"):
                continue
            entry = line.split("\t", 1)[0]
            entries[entry].append(line)
    return entries



# coding: utf-8

from botok import *
import pyewts

from .utils.regex_batch_apply import get_regex_pairs, batch_apply_regex
from .utils.profile_report import profile_report

# from .utils.bo_sorted import bo_sorted
from .pipeline.pipes import pybo_prep, pybo_mod, pybo_form
from .corpus.parse_corrected import parse_corrected, extract_new_entries

__version__ = "0.6.22"



# coding: utf-8
from pathlib import Path
from shutil import copyfile

from bordr import rdr as r

from .rdr_2_replace_matcher import rdr_2_replace_matcher


def rdr_postprocess(rules, infile, outdir=None, keep="model"):
    suffixes = [".DICT", ".INIT", ".RAW", ".RDR", ".sDict"]
    if not outdir:
        outdir = infile.parent.parent
    else:
        outdir = Path(outdir)

    # write adjustment file
    adj_file = outdir / (infile.stem + ".tsv")
    adj_file.write_text(rules, encoding="utf-8-sig")

    # copy files to output directory
    for s in suffixes:
        if keep == "all":
            src = infile.parent / (infile.name + s)
            dst = outdir / (infile.name + s)
            if src != dst:
                copyfile(src, dst)
                Path(infile.parent / (infile.name + s)).unlink()
        elif keep == "model":
            if s in [".DICT", ".RDR"]:
                src = infile.parent / (infile.name + s)
                dst = outdir / (infile.name + s)
                if src != dst:
                    copyfile(src, dst)
                    Path(infile.parent / (infile.name + s)).unlink()
            else:
                Path(infile.parent / (infile.name + s)).unlink()
        elif keep == "none":
            Path(infile.parent / (infile.name + s)).unlink()
        else:
            raise SyntaxError("'keep' should either be 'all', 'model' or 'none'.")


def rdr(infile, outdir=None, keep="model"):
    """

    :param infile: file to process. should be a POS tagged file
    :param outdir: optional. should be the output directory
    :param keep: all RDR files if "all", the .RDR and .DICT files if "model", none if None
    :return: RDR's log
    """
    infile = Path(infile).resolve()

    # run the RDR training
    log = r(str(infile), mode="train", verbose=True)

    # translate to adjustment tsv
    rdr_rules = Path(infile.parent / (infile.name + ".RDR")).read_text(
        encoding="utf-8-sig"
    )
    rules = rdr_2_replace_matcher(rdr_rules)

    # remove RDR files and copy them if needed
    rdr_postprocess(rules, infile, outdir=outdir, keep=keep)

    return log if log else None



# coding: utf-8
from botok import TokChunks, NAMCHE, TSEK


def word_cleanup(string):
    """If it is Tibetan text, returns the cleaned up syllables, otherwise the original string"""

    def join_syls(syls):
        return "".join([syl if syl.endswith(NAMCHE) else syl + TSEK for syl in syls])

    syls = TokChunks(string).get_syls()
    if syls:
        return join_syls(syls)
    else:
        return string



from collections import defaultdict

# variables
tag = "object.tag"
word = "object.word"
prev_tag = "object.prevTag"
next_tag = "object.nextTag"
prev_word = "object.prevWord"
next_word = "object.nextWord"
conclusion = "object.conclusion"
suffix = "object.suffixL"
op = " == "
ccl_op = " = "
cond_sep = " and "
rule_sep = " : "
cql_rule_sep = " & "
level_sep = "\t"

positive = [tag, word, next_tag, next_word, conclusion, suffix]
negative = [prev_tag, prev_word]
eq_table = {
    tag: "pos",
    prev_tag: "pos",
    next_tag: "pos",
    conclusion: "pos",
    word: "text",
    prev_word: "text",
    next_word: "text",
    suffix: "text",
}


def rdr_2_replace_matcher(string):
    cql = format_rules(find_rules(find_levels(string)))
    repl = "{match_cql}\t{replace_index}\t=\t{replace_cql}"
    repls = [
        repl.format(match_cql=a, replace_index=b, replace_cql=c)
        for a, b, c in cql
        if a != c
    ]
    return "\n".join(repls)


def format_rules(rules):
    def generate_cql(test):
        if len(test) > 2:
            s, *_, e = sorted(test)
        elif len(test) == 2:
            s, e = sorted(test)
        else:
            s, e = 0, 0

        slots = []
        slot_zero_idx = None
        for num, t in enumerate(range(s, e + 1)):
            if t == 0:
                slot_zero_idx = num + 1

            if t in test:
                conds = [f"{eq_table[tag]}={pos}" for tag, pos in test[t]]
                slots.append("[" + cql_rule_sep.join(conds) + "]")
            else:
                slots.append("[]")
        assert slot_zero_idx is not None
        return " ".join(slots), slot_zero_idx

    cql = []
    for rule in rules:
        test_cql, idx = generate_cql(rule["test"])
        ccl_cql, _ = generate_cql(rule["ccl"])
        cql.append((test_cql, idx, ccl_cql))
    return cql


def find_levels(string):
    out = []
    for line in string.split("\n"):
        if not line:
            continue
        count = 0
        while line[0] == level_sep:
            count += 1
            line = line[1:]
        out.append((count, line))
    return out


def find_rules(lines):
    rules = []

    # state == {<level/int>: <test>, ...}
    # test == {<position/int>: (<tag>, <POS>), ...}
    state = {}
    for level, line in lines:
        # if level 0, pass. there is no rule to implement
        if level == 0:
            continue

        tests, ccl = parse_line(line)
        ordered_tests = defaultdict(list)
        for t in tests:
            for pos, test in t.items():
                ordered_tests[pos].append(test)

        # save current rule in state to use in indented rules
        state[level] = ordered_tests

        test = defaultdict(list)
        for l in range(1, level + 1):
            for pos, t in state[l].items():
                for u in t:
                    if u not in test[pos]:  # avoid duplicates
                        test[pos].append(u)
        rules.append({"test": test, "ccl": ccl})
    return rules


def parse_line(line):
    rule, ccl = line.split(rule_sep)
    tests = rule.split(cond_sep)
    ccl = parse_test(ccl)
    ccl[0] = [ccl[0]]
    tests = [parse_test(t) for t in tests]
    return tests, ccl


def parse_test(test):
    def parser(test, op):
        pos = 0
        attr, tag = test.split(op)
        for p in positive:
            if p in attr and len(attr) > len(p):
                pos = int(attr[-1])
                attr = attr[:-1]
        for n in negative:
            if n in attr and len(attr) > len(n):
                pos = -int(attr[-1])
                attr = attr[:-1]

        if attr == suffix:
            tag = '".*' + tag[1:]
        return attr, pos, tag

    if op in test:
        attr, pos, tag = parser(test, op)
    elif ccl_op in test:
        attr, pos, tag = parser(test, ccl_op)
    else:
        raise SyntaxError
    return {pos: (attr, tag)}



# coding: utf-8


def pybo_prep(in_str):
    # remove all returns from input
    in_str = in_str.replace("\n", "")
    return in_str


# tokenizing method implemented in cli.py


def get_tag(token, tag_code):
    maps = {"r": "text", "t": "text_cleaned", "p": "pos", "l": "lemma", "s": "sense"}
    try:
        return token[maps[tag_code]]
    except:
        return ""


def pybo_mod(tokens, tag_codes=None):
    """extract text/pos tuples from Token objects"""
    txt_tags = []
    for token in tokens:
        tags = []
        tags.append(token.text)
        # Select and order the tags
        for tag_code in tag_codes:
            tags.append(get_tag(token, tag_code))
        txt_tags.append(tags)
    return txt_tags


def ws2uc(tags):
    """Convert whitespace in raw-text to underscore."""
    tags[0] = tags[0].replace(" ", "_")
    return tags


def pybo_form(tokens, sep=" "):
    """Format in a single string to be written to file"""
    # concatenate text/pos in a single string and replace all in-token spaces by underscores
    out = ["/".join(ws2uc(t)) for t in tokens]
    return sep.join(out)
